@@ -3,6 +3,7 @@
  *
  * Copyright © 2009-2012 Siegfried-A. Gevatter Pujals <rainct@ubuntu.com>
  * Copyright © 2009-2011 Joe Burmeister <joe.a.burmeister@googlemail.com>
+ * Copyright © 2015-2017 asrp <asrp@email.com>
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -21,78 +22,120 @@
 #include <Python.h>
 #include <espeak/speak_lib.h>
 
+#if PY_MAJOR_VERSION < 3
+	#define PyLong_FromLong PyInt_FromLong
+#endif
+
 // Exception prototypes
 static PyObject *BufferFullError;
 static PyObject *CallBack = NULL;
 static volatile int Stopping = 0;
 static volatile int AsyncPython = 0;
 
-#if PY_MAJOR_VERSION >= 3
-	#define PyInt_FromLong PyLong_FromLong
-#endif
+const char* wave_filename = NULL;
+static PyObject* wave_filename_obj = NULL;
 
 static int
-DoCallback(espeak_EVENT_TYPE event, int pos, int len)
+DoCallback(short* wave, espeak_EVENT_TYPE event, int pos, int len, int num_samples, const char* name)
 {
 	int isTrue = 1;
-
+	if (wave != NULL && wave_filename != NULL) {
+		FILE* output = fopen(wave_filename, "w+");
+		fwrite(wave, num_samples*2, 1, output);
+		fclose(output);
+	}
 	PyObject* result = PyObject_CallFunction(
-		CallBack, const_cast<char *>("iii"), event, pos, len);
+		CallBack, const_cast<char *>("iiiis"), event, pos, len, num_samples, name);
 	
-	if (result != NULL)
-	{
+	if (result != NULL) {
 		isTrue = PyObject_IsTrue(result);
 		Py_DECREF(result);
-	}
+	}/* else {
+	    printf("Call failed!\n");
+	}*/
 	
 	return isTrue;
 }
 
 int
-PyEspeakCB(short*, int num, espeak_EVENT* event)
+PyEspeakCB(short* wave, int num_samples, espeak_EVENT* event)
 {
-	if (CallBack != NULL && event != NULL && Stopping == 0)
-	{
+	int count = 0;
+	if (CallBack != NULL && event != NULL && Stopping == 0) {
 		AsyncPython = 1;
-		while (event->type != espeakEVENT_LIST_TERMINATED && Stopping == 0)
-		{
+		while (Stopping == 0) {
+			count++;
 			int isTrue = 1;
 			PyGILState_STATE gs = PyGILState_Ensure();
-			isTrue = DoCallback(
-				event->type, event->text_position, event->length);
+			const char *name = NULL;
+			if (event->type == espeakEVENT_MARK || event->type == espeakEVENT_PLAY) {
+              name = (event->id).name;
+			} else if (event->type == espeakEVENT_SAMPLERATE){
+			  char str[15];
+			  sprintf(str, "%d", event->id.number);
+			  name = str;
+			}
+			if (event->type != espeakEVENT_LIST_TERMINATED) {
+				isTrue = DoCallback(NULL, event->type, event->text_position, event->length, 0, name);
+			} else {
+				isTrue = DoCallback(wave, event->type, event->text_position, event->length, num_samples, name);
+			}
 			PyGILState_Release(gs);
-			if (!isTrue)
-			{
+			if (!isTrue) {
 				AsyncPython = 0;
 				return 1; //abort
 			}
+			if (event->type == espeakEVENT_LIST_TERMINATED) break;
 			++event;
 		}
 		AsyncPython = 0;
 	}
-	
+
 	return Stopping;
 }
 
-static bool
-pyespeak_initialize() {
-	bool result = espeak_Initialize(
-		AUDIO_OUTPUT_PLAYBACK,
-		400,	// sound buffer length (not used)
-		NULL,	// use default data directory
-		0		// no flags
-	);
-	
-	if (result)
-		espeak_SetSynthCallback(PyEspeakCB);
-	
-	return result;
+static PyObject *
+pyespeak_initialize(PyObject *self, PyObject *args, PyObject *kwdict) {
+	int synchronous = 0;
+	int playback = 1;
+	int buffer_length = 400;
+	espeak_AUDIO_OUTPUT espeak_mode;
+	static const char *kwlist[] = {"synchronous", "playback", "buffer_length", NULL};
+	if (!PyArg_ParseTupleAndKeywords(args, kwdict, "|iii",
+		 const_cast<char **>(kwlist),
+		 &synchronous, &playback, &buffer_length))
+	  return NULL;
+
+	if (synchronous && playback) {
+	  espeak_mode = AUDIO_OUTPUT_SYNCH_PLAYBACK;
+	} else if (!synchronous && playback) {
+	  espeak_mode = AUDIO_OUTPUT_PLAYBACK;
+	} else if (synchronous && !playback) {
+	  espeak_mode = AUDIO_OUTPUT_SYNCHRONOUS;
+	} else {
+	  espeak_mode = AUDIO_OUTPUT_RETRIEVAL;
+	}
+	bool result = espeak_Initialize(espeak_mode,
+								    buffer_length,	// sound buffer length
+									NULL,	// use default data directory
+									0		// no flags
+									);
+	if (result) {
+      espeak_SetSynthCallback(PyEspeakCB);
+      Py_INCREF(Py_True);
+      return Py_True;
+    }
+	Py_INCREF(Py_False);
+	return Py_False;
 }
 
 static void
 pyespeak_finalize() {
-	
-	Py_CLEAR(CallBack);
+    //Segfaults if CallBack is an instance method instead of a function.
+	if (CallBack != NULL)
+		Py_CLEAR(CallBack);
+	if (wave_filename_obj != NULL)
+		Py_CLEAR(wave_filename_obj);
 	espeak_Terminate();
 }
 
@@ -100,14 +143,16 @@ static PyObject *
 pyespeak_synth(PyObject *self, PyObject *args, PyObject *kwdict) {
 	int r = EE_OK;
 	const char *text;
+	int start_pos = 0;
+	int end_pos = 0;
 	PyObject* enable_ssml = NULL;
 	PyObject* enable_phonemes = NULL;
 	PyObject* enable_endpause = NULL;
 	PyObject* user_data = NULL;
 	
-	static const char *kwlist[] = {"text", "ssml", "phonemes", "endpause", "user_data", NULL};
-	if(!PyArg_ParseTupleAndKeywords(args, kwdict, "s|OOOO",
-        const_cast<char **>(kwlist), &text,
+	static const char *kwlist[] = {"text", "start_pos", "end_pos", "ssml", "phonemes", "endpause", "user_data", NULL};
+	if (!PyArg_ParseTupleAndKeywords(args, kwdict, "s|iiOOOO",
+        const_cast<char **>(kwlist), &text, &start_pos, &end_pos,
         &enable_ssml, &enable_phonemes, &enable_endpause, &user_data))
 		return NULL;
 	
@@ -121,12 +166,11 @@ pyespeak_synth(PyObject *self, PyObject *args, PyObject *kwdict) {
 	
 	size_t len = strlen(text) + 1;
 	
-	if (len > 0)
-	{
-		r = espeak_Synth(text, len, 0, POS_CHARACTER, 0,
+	if (len > 0) {
+		r = espeak_Synth(text, len, start_pos, POS_CHARACTER, end_pos,
 			flags | espeakCHARS_AUTO, NULL, (void *) user_data);
 		
-		if(r == EE_BUFFER_FULL) {
+		if (r == EE_BUFFER_FULL) {
 			PyErr_SetString(BufferFullError, "command could not be buffered");
 			return NULL;
 		} else if(r == EE_INTERNAL_ERROR) {
@@ -136,32 +180,28 @@ pyespeak_synth(PyObject *self, PyObject *args, PyObject *kwdict) {
 			Py_INCREF(Py_True);
 			return Py_True;
 		}
-	}
-	else
-	{
+	} else {
 		Py_INCREF(Py_False);
 		return Py_False;
 	}
 }
 
 static PyObject *
-pyespeak_set_SynthCallback(PyObject *self, PyObject *args)
+pyespeak_set_synth_callback(PyObject *self, PyObject *args)
 {
 	PyObject* cb;
 	
-	if(!PyArg_ParseTuple(args, "O", &cb))
-	{
+	if (!PyArg_ParseTuple(args, "O", &cb)) {
 		PyErr_SetString(BufferFullError, "invalid argument");
 		return NULL;
 	}
-	
+
+    //What??? Why can't there be more than one callback?
 	if (CallBack != NULL)
 		Py_CLEAR(CallBack);
 	
-	if(Py_None != cb)
-	{
-		if(!PyCallable_Check(cb))
-		{
+	if(Py_None != cb) {
+		if(!PyCallable_Check(cb)) {
 			PyErr_SetString(BufferFullError, "not callable object");
 			return NULL;
 		}
@@ -169,22 +209,18 @@ pyespeak_set_SynthCallback(PyObject *self, PyObject *args)
 		Py_INCREF(cb);
 		CallBack = cb;
 	}
-	
-	Py_INCREF(Py_True);
-	return Py_True;
+
+    Py_RETURN_NONE;
 }
 
 
 static PyObject *
-pyespeak_is_playing(PyObject *self, PyObject *args)
+pyespeak_playing(PyObject *self, PyObject *args)
 {
-    if (espeak_IsPlaying())
-    {
+    if (espeak_IsPlaying()) {
 	    Py_INCREF(Py_True);
 	    return Py_True;
-    }
-    else
-    {
+    } else {
 	    Py_INCREF(Py_False);
 	    return Py_False;
     }
@@ -192,7 +228,7 @@ pyespeak_is_playing(PyObject *self, PyObject *args)
 
 
 static PyObject *
-pyespeak_cancel(PyObject *self, PyObject *args)
+pyespeak_stop(PyObject *self, PyObject *args)
 {
 	Stopping = 1;
     // We need to release the GIL to avoid a deadlock with PyGILState_Ensure.
@@ -203,8 +239,7 @@ pyespeak_cancel(PyObject *self, PyObject *args)
 	PyEval_RestoreThread(ts);
 	Stopping = 0;
 	
-	Py_INCREF(Py_True);
-	return Py_True;
+    Py_RETURN_NONE;
 }
 
 static PyObject *
@@ -215,18 +250,17 @@ pyespeak_set_voice(PyObject *self, PyObject *args, PyObject *kwdict) {
 	voice.gender = 0;
 	voice.age = 0;
 	voice.variant = 0;
-	
+
 	static const char *kwlist[] = {"name", "language", "gender", "age",
 		"variant", NULL};
-	if(!PyArg_ParseTupleAndKeywords(args, kwdict, "|ssiii",
+	if (!PyArg_ParseTupleAndKeywords(args, kwdict, "|ssiii",
         const_cast<char **>(kwlist), &voice.name, &voice.languages,
         &voice.gender, &voice.age, &voice.variant))
 		return NULL;
 	
 	espeak_SetVoiceByProperties(&voice);
 	
-	Py_INCREF(Py_True);
-	return Py_True;
+    Py_RETURN_NONE;
 }
 
 static PyObject *
@@ -238,7 +272,7 @@ pyespeak_set_parameter(PyObject *self, PyObject *args, PyObject *kwdict) {
 	int r;
 	
 	static const char *kwlist[] = {"parameter", "value", "relative", NULL};
-	if(!PyArg_ParseTupleAndKeywords(args, kwdict, "ii|O",
+	if (!PyArg_ParseTupleAndKeywords(args, kwdict, "ii|O",
         const_cast<char **>(kwlist), &parameter, &value, &isRelative))
 		return NULL;
 	
@@ -246,7 +280,7 @@ pyespeak_set_parameter(PyObject *self, PyObject *args, PyObject *kwdict) {
 	
 	r = espeak_SetParameter((espeak_PARAMETER)parameter, value, relative);
 	
-	if(r == EE_BUFFER_FULL) {
+	if (r == EE_BUFFER_FULL) {
 		PyErr_SetString(BufferFullError, "command could not be buffered");
 		return NULL;
 	} else if(r == EE_INTERNAL_ERROR) {
@@ -264,12 +298,12 @@ pyespeak_get_parameter(PyObject *self, PyObject *args) {
 	int current = 0;
 	PyObject* getCurrent;
 	
-	if(!PyArg_ParseTuple(args, "iO", &parameter, &getCurrent))
+	if (!PyArg_ParseTuple(args, "iO", &parameter, &getCurrent))
 		return NULL;
 	
 	current = PyObject_IsTrue(getCurrent);
 	
-	return PyInt_FromLong(espeak_GetParameter((espeak_PARAMETER)parameter, current));
+	return PyLong_FromLong(espeak_GetParameter((espeak_PARAMETER)parameter, current));
 }
 
 static PyObject *
@@ -297,22 +331,51 @@ pyespeak_list_voices(PyObject *self, PyObject *args) {
 	return python_list;
 }
 
+
+static PyObject*
+pyespeak_set_wave_filename(PyObject *self, PyObject *args){
+  // Reading args into "s" gives a char* that is garbage collected after this call.
+  if (wave_filename_obj != NULL)
+    Py_CLEAR(wave_filename_obj);
+  if (!PyArg_ParseTuple(args, "s", &wave_filename)) {
+    PyErr_SetString(BufferFullError, "invalid argument");
+    return NULL;
+  }
+  wave_filename_obj = Py_BuildValue("s", wave_filename);
+  Py_INCREF(wave_filename_obj);
+  //printf("%s\n", wave_filename);
+  Py_RETURN_NONE;
+}
+
+static PyObject* 
+pyespeak_get_wave_filename(PyObject *self, PyObject *args){
+  if (wave_filename_obj != NULL)
+    return wave_filename_obj;
+  Py_RETURN_NONE;
+}
+
 /* Module Methods Table */
 static PyMethodDef EspeakMethods[] = {
+	{"init", (PyCFunction)pyespeak_initialize, METH_VARARGS | METH_KEYWORDS,
+		"Initialization. Should be called before any other function."},
 	{"synth", (PyCFunction)pyespeak_synth, METH_VARARGS | METH_KEYWORDS,
 		"Synthesizes the given text."},
-	{"cancel", pyespeak_cancel, METH_VARARGS,
-		"Stops speech synthesize."},
-	{"is_playing", pyespeak_is_playing, METH_VARARGS,
-		"Queries whether speech synthesize is in progress."},
-	{"set_SynthCallback", pyespeak_set_SynthCallback, METH_VARARGS,
+	{"stop", pyespeak_stop, METH_VARARGS,
+		"Stops speech synthesizer."},
+	{"playing", pyespeak_playing, METH_VARARGS,
+		"Returns True if speech synthesis is in progress."},
+	{"set_synth_callback", pyespeak_set_synth_callback, METH_VARARGS,
 		"Sets a sync callback."},
 	{"set_voice", (PyCFunction)pyespeak_set_voice, METH_VARARGS | METH_KEYWORDS,
-  		"Changes the used voice to one matching the given characteristics."},
+		"Change current voice characteristics."},
 	{"set_parameter", (PyCFunction)pyespeak_set_parameter, METH_VARARGS | METH_KEYWORDS,
 		"Changes a parameter, which may be one of: rate, volume, pitch, range, punctuation, capitals, wordgap."},
 	{"get_parameter", pyespeak_get_parameter, METH_VARARGS,
 		"Retrieves a parameter, which may be one of: rate, volume, pitch, range, punctuation, capitals, wordgap."},
+	{"get_wave_filename", pyespeak_get_wave_filename, METH_VARARGS,
+		"Get the name of the temporary file for sending wave content on callback."},
+	{"set_wave_filename", pyespeak_set_wave_filename, METH_VARARGS,
+		"Set the name of the temporary file for sending wave content on callback."},
 	{"list_voices", pyespeak_list_voices, METH_VARARGS,
 		"Lists all voices."},
 	{NULL, NULL, 0, NULL}
@@ -356,6 +419,7 @@ MOD_INIT(core) {
 	PyModule_AddIntConstant(module, "parameter_WORDGAP", espeakWORDGAP);
 	
 	// Add event types
+	PyModule_AddIntConstant(module, "event_LIST_TERMINATED", espeakEVENT_LIST_TERMINATED);
 	PyModule_AddIntConstant(module, "event_WORD", espeakEVENT_WORD);
 	PyModule_AddIntConstant(module, "event_SENTENCE", espeakEVENT_SENTENCE);
 	PyModule_AddIntConstant(module, "event_MARK", espeakEVENT_MARK);
@@ -363,7 +427,7 @@ MOD_INIT(core) {
 	PyModule_AddIntConstant(module, "event_END", espeakEVENT_END);
 	PyModule_AddIntConstant(module, "event_MSG_TERMINATED", espeakEVENT_MSG_TERMINATED);
 	PyModule_AddIntConstant(module, "event_PHONEME", espeakEVENT_PHONEME);
-	
+	PyModule_AddIntConstant(module, "event_SAMPLE_RATE", espeakEVENT_SAMPLERATE);
 	// Add punctuation types
 	PyModule_AddIntConstant(module, "punctuation_NONE", espeakPUNCT_NONE);
 	PyModule_AddIntConstant(module, "punctuation_ALL", espeakPUNCT_ALL);
@@ -375,12 +439,6 @@ MOD_INIT(core) {
 	Py_INCREF(BufferFullError);
 	PyModule_AddObject(module, "error", BufferFullError);
 	
-	// Initialize eSpeak
-	if(pyespeak_initialize() == -1) {
-		PyErr_SetString(PyExc_SystemError, "could not initialize espeak");
-		return MOD_ERROR_VAL;
-	}
-	
 	// Setup destructor
 	atexit(pyespeak_finalize);
 
